#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from sysrepo_interfaces import utils as utils
from netaddr import IPNetwork
import re
import os
from pyroute2 import IPDB
from platform import machine

rtindex_offset = 1
is_arm_arch = ('aarch64' in machine())

def get_port_fwd():
    lck_f = open('/run/port-fwd.lck', 'a+')
    import fcntl
    import json
    try:
        fcntl.lockf(lck_f.fileno(), fcntl.LOCK_EX)
    except IOError:
        print('Upgrade procedure already running !!!')
        return {}

    if os.path.exists('/config/port-fwd.cfg'):
        jf = open('/config/port-fwd.cfg', 'r+')
        cfg = json.load(jf)
        jf.close()
    else:
        cfg = {}

    fcntl.lockf(lck_f.fileno(), fcntl.LOCK_UN)
    lck_f.close()
    return cfg

def set_nat_iptables(self, nat, wan, wan_tag=None):
    sys_wan = utils.get_ifname(self,wan, vlan=wan_tag)
    for iptables in {'iptables', 'ip6tables'}:
        if 'wifi' in nat:
            utils.shell_return_ok(self,"{} -A FORWARD -i {} -o br-{} -m state --state RELATED,ESTABLISHED -j ACCEPT".format(iptables, sys_wan, nat))
            utils.shell_return_ok(self,"{} -A FORWARD -i br-{} -o {} -j ACCEPT".format(iptables, nat, sys_wan))
        else:
            utils.shell_return_ok(self,"{} -A FORWARD -i {} -o {} -m state --state RELATED,ESTABLISHED -j ACCEPT".format(iptables, sys_wan, nat))
            utils.shell_return_ok(self,"{} -A FORWARD -i {} -o {} -j ACCEPT".format(iptables, nat, sys_wan))

        utils.shell_return_ok(self,"{} -t nat -A POSTROUTING -o {} -j MASQUERADE".format(iptables, sys_wan))

    if 'wifi' not in nat:
        port_fwd = get_port_fwd()
        if nat in port_fwd.keys():
            port_fwd_lst = port_fwd[nat]
            for entry in port_fwd_lst:
                dst_addr_ver = IPNetwork(entry['dst-addr']).version
                if dst_addr_ver == 6:
                    iptables = "ip6tables"
                else:
                    iptables = "iptables"

                ipt_cmd = '{iptables} -t nat -A PREROUTING -i {wan_if} -p {proto} --dport {port} -j DNAT --to {dst_addr}:{dst_port}'.format(iptables=iptables, wan_if=sys_wan, proto=entry['proto'],
                                                                                                                                            port=entry['port'], dst_addr=entry['dst-addr'], dst_port=str(entry['dst-port']))
                utils.shell_return_ok(self,ipt_cmd)

def is_nat_change(self, change):
    name_element = change['xpath'].lstrip(self.namespace).split('/', 1)
    if name_element:
        r = re.compile(".*nat[0-9]{1,2}.*")
        m = r.match(name_element[0])
        if m is None:
            return False
        else:
            return True
    return False

def clean_wan_routes(self, nat, wan, wan_tag=None):
    global rtindex_offset
    sys_wan = utils.get_ifname(self,wan, vlan=wan_tag)
    if 'wifi' in nat:
        if 'vap' in nat:
            table_id = int(nat.split("_")[0].replace('wifi','')) + rtindex_offset + 5 + int(nat.split("_")[1].replace("vap",""))
        else:
            table_id = int(nat.replace('wifi', '')) + rtindex_offset + 5
    else:
        table_id = int(nat.replace('nat', '')) + rtindex_offset

    ipdb = IPDB()
    if sys_wan in ipdb.interfaces:
        wan_idx = ipdb.interfaces[sys_wan]
        if table_id in ipdb.routes.tables.keys():
            for rt in ipdb.routes.tables[table_id]:
                dst = "none"
                if 'dst' in rt:
                    dst = rt['dst']
                if rt['oif'] == wan_idx['index']:
                    try:
                        with ipdb.routes.tables[str(table_id)][rt['dst']] as r:
                            r.remove()
                    except:
                        self.logger.error("Can't delete route %s from table %s" % (dst,str(table_id)))
                if os.path.isfile("/config/nat-lb"):
                    os.system("ip route del  default  metric 1  table %d" % table_id)

def create_wifi_nat_interface(self, name, config, change_set):
    global rtindex_offset

    if 'nat' in config and config['nat']:
        if 'wifi-wan-interface' in config:
            for wan in config['wifi-wan-interface']:
# TODO: revisit with multi VLANs
#                vlan = self.get_interface_param(wan['name'], 'vlan', None)
                vlan = None
                set_nat_iptables(self, name, wan['name'], wan_tag=vlan)
                set_wan_routes(self, name, wan['name'], wan_tag=vlan)
                if self.natpassthr is not None:
                    self.logger.debug('Add wan interface \'{}\' to pass-thr for nat \'br-{}\' ' .format(wan['name'], name))
                    self.natpassthr.nat_wan_add("br-{}".format(name), wan['name'])
        if 'vap' in name:
            table_id = int(name.split("_")[0].replace('wifi','')) + rtindex_offset + 5 + int(name.split("_")[1].replace("vap",""))
        else:
            table_id = int(name.replace('wifi', '')) + rtindex_offset + 5
        self.execute_shell_command("ip rule add iif br-{} unreachable".format(name))
        self.execute_shell_command("ip rule add iif br-{} table {}".format(name, table_id))

def clean_nat_iptables(self, nat, wan, wan_tag=None):
    sys_wan = utils.get_ifname(self,wan, vlan=wan_tag)
    for iptables in {'iptables', 'ip6tables'}:
        if 'wifi' in nat:
            utils.shell_return_ok(self,"{} -D FORWARD -i {} -o br-{} -m state --state RELATED,ESTABLISHED -j ACCEPT".format(iptables, sys_wan, nat))
            utils.shell_return_ok(self,"{} -D FORWARD -i br-{} -o {} -j ACCEPT".format(iptables, nat, sys_wan))
        else:
            utils.shell_return_ok(self,"{} -D FORWARD -i {} -o {} -m state --state RELATED,ESTABLISHED -j ACCEPT".format(iptables, sys_wan, nat))
            utils.shell_return_ok(self,"{} -D FORWARD -i {} -o {} -j ACCEPT".format(iptables, nat, sys_wan))

        utils.shell_return_ok(self,"{} -t nat -D POSTROUTING -o {} -j MASQUERADE".format(iptables, sys_wan))
        utils.shell_return_ok(self,"conntrack  -D -n")

    if 'wifi' not in nat:
        port_fwd = get_port_fwd()
        if nat in port_fwd.keys():
            port_fwd_lst = port_fwd[nat]
            for entry in port_fwd_lst:
                dst_addr_ver = IPNetwork(entry['dst-addr']).version
                if dst_addr_ver == 6:
                    iptables = "ip6tables"
                else:
                    iptables = "iptables"

                ipt_cmd = '{iptables} -t nat -D PREROUTING -i {wan_if} -p {proto} --dport {port} -j DNAT --to {dst_addr}:{dst_port}'.format(iptables=iptables, wan_if=sys_wan, proto=entry['proto'],
                                                                                                                                            port=entry['port'], dst_addr=entry['dst-addr'], dst_port=str(entry['dst-port']))
                utils.shell_return_ok(self,ipt_cmd)

def delete_nat_interface(self, name, config):
    global rtindex_offset
    # Remove IPTable config
    if 'wan-interface' in config:
        for wan in config['wan-interface']:
# TODO: revisit with multi VLANs
#            wan_tag = self.get_interface_param(wan['name'], 'vlan', None)
            wan_tag = None
            clean_nat_iptables(self, name, wan['name'], wan_tag=wan_tag)
    if 'wifi-wan-interface' in config:
        for wan in config['wifi-wan-interface']:
# TODO: revisit with multi VLANs
#            wan_tag = self.get_interface_param(wan['name'], 'vlan', None)
            wan_tag = None
            clean_nat_iptables(self, name, wan['name'], wan_tag=wan_tag)

    if 'wifi' in name:
        if '_vap' in name:
            table_id = int(name.split("_")[0].replace('wifi','')) + rtindex_offset + 5 + int(name.split("_")[1].replace("vap",""))
        else:
            table_id = int(name.replace('wifi', '')) + rtindex_offset + 5
        utils.shell_return_ok(self,"ip rule del iif br-{} table {}".format(name, table_id))
        utils.shell_return_ok(self,"ip rule del iif br-{} unreachable".format(name))
        utils.shell_return_ok(self,"ip route flush table {}".format(table_id))
        utils.shell_return_ok(self,"conntrack  -D -n")
    else:
        table_id = int(name.replace('nat', '')) + rtindex_offset
        utils.shell_return_ok(self,"ip rule del iif {} table {}".format(name, table_id))
        utils.shell_return_ok(self,"ip rule del iif {} unreachable".format(name))
        utils.shell_return_ok(self,"ip route flush table {}".format(table_id))
        utils.shell_return_ok(self,"conntrack  -D -n")
        self.execute_shell_command("/scripts/apply_networking.py -z {}".format(name))

def set_wan_routes(self, nat, wan, wan_tag=None):
    global rtindex_offset
    global is_arm_arch
    sys_wan = utils.get_ifname(self, wan, vlan=wan_tag)
    if 'wifi' in nat:
        if 'vap' in nat:
            table_id = int(nat.split("_")[0].replace('wifi','')) + rtindex_offset + 5 + int(nat.split("_")[1].replace("vap",""))
        else:
            table_id = int(nat.replace('wifi', '')) + rtindex_offset + 5
    else:
        table_id = int(nat.replace('nat', '')) + rtindex_offset

    ipdb = IPDB()
    from pyroute2.netlink.exceptions import NetlinkError
    if sys_wan in ipdb.interfaces:
        wan_idx = ipdb.interfaces[sys_wan]
        for rt in ipdb.routes:
            # ARM has problem with iproute2, it returns IPv4 routes as IPV6 ones
            if is_arm_arch and rt['family'] == 10:
                continue
            if rt['oif'] == wan_idx['index']:
                if rt['gateway']:
                    if rt['gateway'].startswith("fe80:") : #link-local address
                        route = {'dst': rt['dst'], 'gateway': rt['gateway'], 'table': table_id, 'priority': rt['priority'], 'oif': wan_idx['index']}
                    else:
                        route = {'dst': rt['dst'], 'gateway': rt['gateway'], 'table': table_id, 'priority': rt['priority']}
                else:
                    route = {'dst': rt['dst'], 'table': table_id, 'oif': wan_idx['index']}

                try:
                    ipdb.routes.add(route).commit()
                except NetlinkError as e:
                    self.logger.error('Set route in table {tbl} failed. {msg}'.format(tbl=table_id, msg=str(e)))
                    if e.code != 17:
                        self.logger.error('Set route in table {tbl} failed. {msg}'.format(tbl=table_id, msg=str(e)))
                if os.path.isfile("/config/nat-lb"):
                    os.system("ip route del  default  metric 1  table %d" % table_id)
                    cmd='ip route add default proto static metric 1 table %d $(for metric in `seq 2 249` ; do  ip route list default table %d metric $metric ; done | sed \'s/default/nexthop/\'  | sed \'s/proto static/weight 1/\' | tr "\n" " ")' %  (table_id, table_id)
                    os.system(cmd)

    ipdb.release()

def create_nat_interface(self, name, config, change_set):
    global rtindex_offset

    if 'enable' in config and config['enable']:
        if 'wan-interface' in config:
            for wan in config['wan-interface']:
# TODO: revisit with multi VLANs
#                vlan = self.get_interface_param(wan['name'], 'vlan', None)
                vlan = None
                set_nat_iptables(self, name, wan['name'], wan_tag=vlan)
                set_wan_routes(self, name, wan['name'], wan_tag=vlan)
                if self.natpassthr is not None:
                    self.logger.debug('Add wan interface \'{}\' to pass-thr for nat \'{}\' ' .format(wan['name'], name))
                    self.natpassthr.nat_wan_add(name, wan['name'])

        table_id = int(name.replace('nat', '')) + rtindex_offset
        self.execute_shell_command("ip rule add iif {} unreachable".format(name))
        self.execute_shell_command("ip rule add iif {} table {}".format(name, table_id))

    ipv4_prefix = self.get_interface_param(name, 'ipv4-prefix', '')
    ipv6_prefix = self.get_interface_param(name, 'ipv6-prefix', '')
    if change_set:
        for xpath, change in change_set.items():
            if change['operation'] == 'DELETED':
                # Fill defaults
                if xpath.find('ipv4-prefix') > -1:
                    ipv4_prefix = ''
                if xpath.find('ipv6-prefix') > -1:
                    ipv6_prefix = ''
                continue
            if xpath.find('ipv4-prefix') > -1:
                ipv4_prefix = change['new_value']
            if xpath.find('ipv6-prefix') > -1:
                ipv6_prefix = change['new_value']

    bootproto = 'manual'
    if ipv4_prefix or ipv6_prefix:
        bootproto = "static"

    options = ["-n {}".format(name), "--ovs-master ovs"]
    if bootproto:
        options.append("-b {}".format(bootproto))
    if ipv4_prefix:
        options.append("--ipv4-prefix={}".format(ipv4_prefix))
    if ipv6_prefix:
        options.append("--ipv6-prefix={}".format(ipv6_prefix))
        # Configure and enable radvd for IPv6
        self.radvd_ifs[name] = ipv6_prefix
    else:
        try:
            if name in self.radvd_ifs.keys():
                del self.radvd_ifs[name]
        except:
            self.logger.info(f'Cannot delete radvd config for: {name}')

    self.execute_shell_command("/scripts/apply_networking.py {}".format(" ".join(options)))

def config_nat_interface(self, name, config, change_set):
    old_enable_nat = False
    if config and 'enable' in config:
        old_enable_nat = config['enable']

    if not config:
        config = {}
    if 'wan-interface' not in config:
        config['wan-interface'] = []
    if change_set:
        for xpath, change in change_set.items():
            if xpath.find('enable') > -1 and change['operation'] != 'DELETED':
                config['enable'] = change['new_value']
                if change['new_value'] and 'true' in str(change['new_value']).lower():
                    config['enable'] = True
                else:
                    config['enable'] = False
            if xpath.find('wan-interface') > -1:
                if change['operation'] == 'DELETED':
                    if not change['old_value']:
                        continue
                    index = 0
                    for wan in config['wan-interface']:
                        if wan['name'] == change['old_value']:
# TODO: revisit with multi VLANs
#                            wan_tag = self.get_interface_param(wan['name'], 'vlan', None)
                            wan_tag = None
                            clean_nat_iptables(self, name, wan['name'], wan_tag=wan_tag)
                            config['wan-interface'].pop(index)
                            if self.natpassthr is not None:
                                self.logger.debug('Delete wan interface \'{}\' from pass-thr for nat \'{}\' ' .format(wan['name'], name))
                                self.natpassthr.nat_wan_del(name, wan['name'])
                            break
                        index = index + 1
                else:
                    if not change['new_value']:
                        continue
                    config['wan-interface'].append({'name': change['new_value']})
            if xpath.find('port-forwarding-all') > -1 and change['operation'] != 'DELETED':
                config['port-forwarding-all'] = change['new_value']
                if change['new_value'] and 'true' in str(change['new_value']).lower():
                    config['port-forwarding-all'] = True
                else:
                    config['port-forwarding-all'] = False
    if old_enable_nat:
        utils.shell_return_ok(self,"ifdown --ignore-error {}".format(name))
        delete_nat_interface(self, name, config)

    create_nat_interface(self, name, config, change_set)
    if 'enable' in config and config['enable']:
        utils.shell_return_ok(self,"ifup --ignore-error {}".format(name))

    if config and 'enable' in config and self.natpassthr is not None:
        self.logger.debug('Update nat interface state \'{}\' to pass-thr for nat \'{}\' ' .format(config['enable'], name))
        self.natpassthr.nat_enable(name, config['enable'])

        if 'port-forwarding-all' in config:
            self.logger.debug('Update pass-thr state \'{}\' to pass-thr for nat \'{}\' ' .format(config['port-forwarding-all'], name))
            self.natpassthr.nat_passthr(name, config['port-forwarding-all'])

def is_wifi_nat_change(self, change):
    name_element = change['xpath'].lstrip(self.namespace).split('/', 1)
    if name_element:
        r = re.compile(".*wifi[0-9]{1,2}.*")
        m = r.match(name_element[0])
        if m is None:
            return False
        else:
            return True
    return False

def apply_wifi_nat_changes(self, change_set, nats):
    nat_change_set = filter(lambda elem: is_wifi_nat_change(self, elem), change_set)
    if not nat_change_set:
        return

    for nat_change in nat_change_set:
        if_nat = utils.get_ifname_from_change(self, nat_change)

        # is wan changed
        change_path = nat_change['xpath'].lstrip(self.namespace).split('/', 1)
        if len(change_path) > 1:
            if change_path[1].startswith('wifi-wan-interface'):
                if nat_change['operation'] == 'CREATED':
                    if nat_change['new_value']:
                        nat_to_ch = [x for x in nats if x['name'] == if_nat]
                        if nat_to_ch:
                            nat_wan_list = nat_to_ch[0].get('wifi-wan-interface', [])
                            nat_wan_list.append({'name': nat_change['new_value']})
                            nat_to_ch[0]['wifi-wan-interface'] = nat_wan_list

                if nat_change['operation'] == 'DELETED':
                    if nat_change['old_value']:
                        nat_to_ch = [x for x in nats if x['name'] == if_nat]
                        if nat_to_ch:
                            nat_wan_list = nat_to_ch[0].get('wifi-wan-interface', [])
                            if nat_wan_list:
                                nat_wan_list.remove({'name': nat_change['old_value']})

def wan_handler(self, change_set):
    running = self.get_running_config(full_path=False)
    for iftype in ['nat', 'wifi']:
        if 'nat' in iftype:
            nats = filter(lambda elem: elem['type'] == 'nat', running)
            apply_nat_changes(self, change_set, nats)
            # self.logger.info(nats)
            wans = filter(lambda elem: ((elem['type'] == 'ethernet' or elem['type'] == 'lte' or elem['type'] == 'wifi' or elem['type'] == 'pppoe') and elem['name'] != 'mgmt0'), running)
        if 'wifi' in iftype:
            nats = filter(lambda elem: elem['type'] == 'wifi', running)
            apply_wifi_nat_changes(self, change_set, nats)
            # self.logger.info(nats)
            wans = filter(lambda elem: ((elem['type'] == 'ethernet' or elem['type'] == 'lte') and elem['name'] != 'mgmt0'), running)

        for change_elem in change_set:
            if_name = utils.get_ifname_from_change(self, change_elem)
            if if_name not in [x['name'] for x in wans]:
                # This interface cannot be wan - skip
                # self.logger.info('Skip {} - now WAN'.format(if_name))
                continue

# TODO: revisit with multi VLANs
#            # self.logger.info('Change for interface {}'.format(if_name))
#            path = change_elem['xpath'].lstrip(self.namespace).split('/')
#            path.pop(0)
#            if len(path) > 0 and path[0] == 'vlan':
#                if 'nat' in iftype:
#                    wan_in = filter(lambda elem: elem.get('wan-interface', None) is not None and if_name in [x['name'] for x in elem.get('wan-interface', None)],nats)
#                if 'wifi' in iftype:
#                    wan_in = filter(lambda elem: elem.get('wifi-wan-interface', None) is not None and if_name in [x['name'] for x in elem.get('wifi-wan-interface', None)],nats)
#
#                # self.logger.info(wan_in)
#                for nat in wan_in:
#                    old_tag = None
#                    new_tag = None
#                    if change_elem['operation'] == 'MODIFIED':
#                        old_tag = int(change_elem['old_value'])
#                        new_tag = int(change_elem['new_value'])
#
#                    if change_elem['operation'] == 'DELETED':
#                        old_tag = int(change_elem['old_value'])
#                        new_tag = None
#
#                    if change_elem['operation'] == 'CREATED':
#                        old_tag = None
#                        new_tag = int(change_elem['new_value'])
#
#                    clean_nat_iptables(self, nat['name'], if_name, wan_tag=old_tag)
#                    clean_wan_routes(self, nat['name'], if_name, wan_tag=old_tag)
#                    set_nat_iptables(self, nat['name'], if_name, wan_tag=new_tag)
#                    set_wan_routes(self, nat['name'], if_name, wan_tag=new_tag)

def apply_nat_changes(self, change_set, nats):
    nat_change_set = filter(lambda elem: is_nat_change(self, elem), change_set)
    if not nat_change_set:
        return

    for nat_change in nat_change_set:
        if_nat = utils.get_ifname_from_change(self, nat_change)

        # is wan changed
        change_path = nat_change['xpath'].lstrip(self.namespace).split('/', 1)
        if len(change_path) > 1:
            if change_path[1].startswith('wan-interface'):
                if nat_change['operation'] == 'CREATED':
                    if nat_change['new_value']:
                        nat_to_ch = [x for x in nats if x['name'] == if_nat]
                        if nat_to_ch:
                            nat_wan_list = nat_to_ch[0].get('wan-interface', [])
                            nat_wan_list.append({'name': nat_change['new_value']})
                            nat_to_ch[0]['wan-interface'] = nat_wan_list

                if nat_change['operation'] == 'DELETED':
                    wan_if = self.get_item(nat_change['xpath'])
                    if wan_if is not None:
                        nat_to_ch = [x for x in nats if x['name'] == if_nat]
                        if nat_to_ch:
                            nat_wan_list = nat_to_ch[0].get('wan-interface', [])
                            if nat_wan_list:
                                nat_wan_list.remove({'name': wan_if})
